//      NOCUDA.c - -  CUDA.cu,    CPU-   CUDA
//
//      Copyright 2011 Edward V. Emelianoff <eddy@sao.ru>
//
//      This program is free software; you can redistribute it and/or modify
//      it under the terms of the GNU General Public License as published by
//      the Free Software Foundation; either version 2 of the License, or
//      (at your option) any later version.
//
//      This program is distributed in the hope that it will be useful,
//      but WITHOUT ANY WARRANTY; without even the implied warranty of
//      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//      GNU General Public License for more details.
//
//      You should have received a copy of the GNU General Public License
//      along with this program; if not, write to the Free Software
//      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
//      MA 02110-1301, USA.

#include "NOCUDA.h"
#include "CUtools.h"
#include <pthread.h>


static float *Sin_d = NULL, *Cos_d = NULL;
static int sincosize = 0;

#ifndef THREAD_NUMBER
	#define THREAD_NUMBER 2		//   - 2 
#endif

//       [-90,+180)
void init_sincos(int angles){
	float anglestep = (float)angles / 270.; //   
	float conv = M_PI / 180.;
	float *cp, *sp;
	int k;
	if(!Sin_d || !Cos_d || angles != sincosize){ //     
		free(Sin_d);
		free(Cos_d);
		Sin_d = malloc(angles * sizeof(float));
		Cos_d = malloc(angles * sizeof(float));
		cp = Cos_d; sp = Sin_d;
		for(k = 0; k < angles; k++){ //      -90  +90  
			float theta = ((float)k / anglestep - 90.) * conv;
			*sp++ = sinf(theta);
			*cp++ = cosf(theta);
			//sincosf(theta, sp++, cp++);
		}
		sincosize = angles;
	}
}

typedef struct{
	float *image;
	float *hough;
	int w;
	int Y0;
	int Y1;
	int Rmax;
	int angles;
	pthread_mutex_t *mutex;
}Hough_kernel_args;
/*
 *      
 */
void *fill_lin_hough_array(void *data){
	Hough_kernel_args *HD = (Hough_kernel_args*) data;
	float *ima = HD->image, *hough = HD->hough;
	pthread_mutex_t *mutex = HD->mutex;
	int Y0 = HD->Y0, Y1 = HD->Y1, imW = HD->w, Rmax = HD->Rmax, angles = HD->angles;
	int i, j, k, Y, R;
	for(j = Y0; j < Y1; j++){
		for(i = 0.; i < imW; i++, ima++){
			if(*ima > 0.1){
				Y = 0;
				pthread_mutex_lock(mutex);
				for(k = 0; k < angles; k++, Y+=Rmax){
					// R = x*cos(theta) + y*sin(theta)
					R = (int)(0.5 + i*Cos_d[k] + j*Sin_d[k]);
					if(R > 0 && R < Rmax) hough[R + Y] += *ima;
				}
				pthread_mutex_unlock(mutex);
			}
		}
	}
	return NULL;
}
/*
 *      
 * :
 * 		ima -  
 * 		imW, imH -    
 * 		Rmax -   R
 * 		angles -    (    180/angles )
 * : hough -    ,
 * 		     
 * 		!!!         
 *      
 */
int fill_hough_lines(float *ima, int imW, int imH, int Rmax, int angles, float *hough){
	int i, Y0, Y1, dY;
	float *hptr = hough, max;
	Hough_kernel_args HD[THREAD_NUMBER];
	pthread_t threads[THREAD_NUMBER];
	pthread_mutex_t mutex;
	pthread_mutex_init(&mutex, NULL);
	init_sincos(angles);
	dY = (imH + THREAD_NUMBER/2) / THREAD_NUMBER;
	Y0 = 0; Y1 = dY;
	for(i = 0; i < THREAD_NUMBER; i++, Y0+=dY, Y1+=dY){
		if(Y0 >= imW) break;
		if(Y1 > imW) Y1 = imW;
		HD[i].image = ima; HD[i].hough = hough;
		HD[i].Y0 = Y0; HD[i].Y1 = Y1; HD[i].w = imW;
		HD[i].Rmax = Rmax; HD[i].angles = angles;
		HD[i].mutex = &mutex;
		pthread_create(&threads[i], NULL, fill_lin_hough_array, &HD[i]);
	}
	for(i = 0; i < THREAD_NUMBER; i++)
		pthread_join(threads[i], NULL);
	pthread_mutex_destroy(&mutex);
	max = *hough++;
	Y1 = Rmax * angles;
	for(i = 1; i < Y1; i++, hough++)
		if(*hough > max) max = *hough;
	hough = hptr;
	for(i = 0; i < Y1; i++, hough++)
		*hough /= max;
	return 1;
}

/*  
int fill_hough_lines(float *ima, int imW, int imH, int Rmax, int angles, float *hough){
	int i, j, k, Y, R;
	float max, *hptr = hough;
	init_sincos(angles);
	for(j = 0.; j < imH; j++){
		for(i = 0.; i < imW; i++, ima++){
			if(*ima > 0.1){
				Y = 0;
				for(k = 0; k < angles; k++, Y+=Rmax){
					// R = x*cos(theta) + y*sin(theta)
					R = (int)(0.5 + i*Cos_d[k] + j*Sin_d[k]);
					if(R > 0 && R < Rmax) hough[R + Y] += *ima;
				}
			}
		}
	}
	max = *hough++;
	Y = Rmax * angles;
	for(i = 1; i < Y; i++, hough++)
		if(*hough > max) max = *hough;
	hough = hptr;
	for(k = 0; k < Y; k++, hough++)
		*hough /= max;
	return 1;
}
*/
