#ifndef _FITSVIEW_H_
#define _FITSVIEW_H_
#ifndef _LARGEFILE64_SOURCE
	#define _LARGEFILE64_SOURCE
#endif
#include <stdio.h>
#include <sys/types.h>
#include <err.h>
#include <stdlib.h>
#include <string.h>
#ifndef  _XOPEN_SOURCE
	#define  _XOPEN_SOURCE 666 // atan2
#endif
#ifndef __USE_XOPEN
	#define __USE_XOPEN		// M_PI
#endif
#include <math.h>
#include <GL/gl.h>
#include <gtk/gtk.h>
#include <gtk/gtkgl.h>
#include <glib.h>
#include <glib/gprintf.h>
#include <libintl.h>
#ifndef __USE_GNU
	#define __USE_GNU
#endif
#include <unistd.h>

//  GPU-    #include CUINCLUDE
#ifdef CUDA_FOUND
	#define CUINCLUDE "CUDA.h"
#else
	#define CUINCLUDE "NOCUDA.h"
#endif

#ifndef GETTEXT_PACKAGE
#define GETTEXT_PACKAGE "fitsview"
#endif
#define PRGNAME GETTEXT_PACKAGE
#ifndef LOCALEDIR
#define LOCALEDIR "./locale"
#endif

#define _(String)				gettext(String)
#define gettext_noop(String)	String
#define N_(String)				gettext_noop(String)

extern const int HIST_SIZE;
void *copy_struct(void *src, int size);
void *init_struct(int size);
#define INIT(var, type)		var = (type*)init_struct(sizeof(type))
#define COPY(src, type)		(type*)copy_struct((void*)src, sizeof(type))

/*            ,
 * ,      */
#define PT_TRESHOLD		5

//  , -DEBUG
#ifdef EBUG
	#define RED			"\033[1;32;41m"
	#define GREEN		"\033[5;30;42m"
	#define OLDCOLOR	"\033[0;0;0m"
	#define FNAME() fprintf(stderr, "\n%s (%s, line %d)\n", __func__, __FILE__, __LINE__)
	#define DBG(...) do{fprintf(stderr, "%s (%s, line %d): ", __func__, __FILE__, __LINE__); \
					fprintf(stderr, __VA_ARGS__);			\
					fprintf(stderr, "\n");} while(0)
#else
	#define FNAME()	 do{}while(0)
	#define DBG(...) do{}while(0)
#endif //EBUG

//  
typedef struct{
	int size;		//  ,      
	int max;		//     
	int min;
	int bot_tres;	//      
	int top_tres;
	int *data;
} ImageHistogram;
//   
struct ImageStat{
	GLfloat max;
	GLfloat min;
	ImageHistogram histogram;
	/*GLfloat median;
	GLfloat mean;
	GLfloat std;*/
};

//  
typedef struct{
	unsigned char drawingMode;	//  : ,   ..
	unsigned char trackingMode;	//   
	unsigned char graphYaxis;	//   Y : ,  
	unsigned char graphMode;	//    (,   ..)
	unsigned char transformMask;//  ,   image_transformed
	unsigned char visualMode;	//   (.   ..)
} Context;

typedef struct{
	unsigned char previewMode;	//   
	gboolean add_all;			//      (  .fts/.fit[s])
} Global_context;

extern Global_context *Global;

/*
 *    
 *  ()     
 */
// drawingMode:
enum{
	 DRAW_TRACK		//    
	,DRAW_SELECTION	//   
	,DRAW_QUAD		//   
	,GAME_MODE		//    3D-   
};
// trackingMode:
enum{
	 TRACK_ANY		//  
	,TRACK_VERT		//  
	,TRACK_HORIZ	//  
	,TRACK_DIAG		//   45
};
// graphYaxis,  :   - ,  - 
enum{
	 Y_LINEAR  = 0	//    ,   
	,Y_LOGBOTH = 3	//    
	,Y_LOGHIST = 2	//   
	,Y_LOGGRAPH= 1	//   
};
// graphMode
enum{
	 GR_GRAPH		//    
	,GR_HISTOGRAM	//    
};
// transformMask -  
enum{
	 TRANS_NONE		= 0  //   
	,TRANS_TRES		= 1  //   (  )
	,TRANS_ROT		= 2  // 
	,TRANS_MOVE		= 4  // 
	,TRANS_SCALE	= 8  // 
	,TRANS_FILTER	= 16 // 
	//   
};
// visualMode -  
enum{
	 SHOW_ONLY_IMAGE	= 0 	//   
	,SHOW_POTBOXES		= 1 	//     
	,SHOW_POTSELECT		= 2 	// .    
	,SHOW_ALL			= 255	//  
};
// previewMode -  , 
enum{
	 PREVIEW_DEFAULT			//   (512512, . )
	,PREVIEW_128				//  
	,PREVIEW_256				// (  )
	,PREVIEW_512
	,PREVIEW_768
	,PREVIEW_1024
	,PREVIEW_SIZEMASK	= 7		//    
	,PREVIEW_LINEAR		= 8		//   
	,PREVIEW_LOG		= 16	//   
	,PREVIEW_SQRT		= 24
	,PREVIEW_COLFNMASK	= 24	//      
};

// 
typedef struct{
	GLint width;		// 
	GLint height;		//   
	GLfloat bZero;		//   
	GLfloat bScale;		// 
	GLfloat maxVal;		//    (1<<BITTPIX)
	int dtype;			//  
	struct ImageStat stat; //   
	GLfloat *data;		//    
	GtkListStore *store;//    
	gchar *filename;	//   (basename)  
	int imagetype;		//   (  )
} IMAGE;
// imagetype
enum{
	 SIMPLE_IMAGE		//  ( ) 
	,HOUGH				//  
};

double dtime();

#include "gtk.h"
#endif // _FITSVIEW_H_
