//      spots.c -     ,    
//
//      Copyright 2011 Edward V. Emelianoff <eddy@sao.ru>
//
//      This program is free software; you can redistribute it and/or modify
//      it under the terms of the GNU General Public License as published by
//      the Free Software Foundation; either version 2 of the License, or
//      (at your option) any later version.
//
//      This program is distributed in the hope that it will be useful,
//      but WITHOUT ANY WARRANTY; without even the implied warranty of
//      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//      GNU General Public License for more details.
//
//      You should have received a copy of the GNU General Public License
//      along with this program; if not, write to the Free Software
//      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
//      MA 02110-1301, USA.

/*
 *     -      
 *         ..,
 *    "" .
 * ,       "".
 *
 */
#include "spots.h"
#include "gtk.h"
#include "gauss.h"

const double M_2PI = 2 * M_PI;
//    11.25  -       :
const double PHI_STEP = 0.05;
//    ( ), 56.25 
const double DPHI_MARKERS = 0.98;
const double aStep = M_PI / 16; // 11.25 ,    
const double aZero = -M_PI + M_PI / 16 * 2.5; //     (  )
const int N_ANGLES = 16; //     ( - )
const int N_RINGS = 8;	//    

//    Y    Y 
double SlopeAngle;
//    (  )   , -101.25
const double Slope0 = -1.767146;

/*
 *   -    
 * !!!     (..  Y   )
 * ,      
 * ,       
 */
double AxisX = 0., AxisY = 0.;


/*
 *   ,     
 */
enum{
	 TOO_MANY_MARKERS			//    
	,NOT_ENOUGH_MARKERS			//    
	,BAD_MARKERS_ANGLE			//     DPHI_MARKERS +- PHI_STEP
	,TOO_MANY_RINGS				//   ,   
};

//   
char *errmsg[] = {
	 N_("Too many markers")
	,N_("Not enough markers")
	,N_("Bad markers angle")
	,N_("Too many rings")
};

#ifdef LEPTONICA_FOUND
/*
 *  -    
 *  ,      
 *   
 */
void bad_circumstances(int errcode){
	DBG(RED "\n\nerror, need to stop\n\n" OLDCOLOR);
	g_err(_(errmsg[errcode]));
}

/*
 *     n 
 */
Spots* spots_alloc(int n){
	Spots *ptr = malloc(sizeof(Spots));
	if(!ptr){
		g_err(_("Can't allocate memory for new spots"));
		return NULL;
	}
	ptr->spot = malloc(n * sizeof(Spot*));
	if(!ptr->spot){
		g_err(_("Can't allocate memory for pointers in spots array"));
		free(ptr);
		return NULL;
	}
	ptr->size = n;
	ptr->n = 0;
	/*    BY_COPY,     
	 * BY_PTR,     free     */
	ptr->memflag = BY_COPY;
	return ptr;
}

void spots_free(Spots **spots){
	int i, n;
	if(!(*spots)) return;
	n = (*spots)->n;
	Spot **spot = (*spots)->spot;
	if((*spots)->memflag != BY_PTR) // ,   spot-  malloc'
		for(i = 0; i < n; i++, spot++)
			free(*spot);
	free((*spots)->spot);
	free(*spots);
	*spots = NULL;
}

/*
 *    spot   spots.
 *     -  NULL
 *
 */
Spot* spots_add(Spots *spots, Spot *spot, int copy){
	Spot *ptr = NULL;
	if(spots->size <= spots->n){
		g_err(_("Spots array is full, can't place new spot\n"));
		return NULL; //  
	}
	if(!spot){
		g_err(_("Zero pointer to spot\n"));
		return NULL;
	}
	if(copy == BY_COPY){
		ptr = malloc(sizeof(Spot));
		if(!ptr){
			g_err(_("Can't allocate memory for new spot"));
			return NULL;
		}
		if(!memcpy(ptr, spot, sizeof(Spot))){
			free(ptr);
			g_err(_("Can't copy spot to new one"));
			return NULL;
		}
	}
	else if(copy == BY_PTR){
		spots->memflag = BY_PTR; //  ,    free
		ptr = spot;
	}
	spots->spot[spots->n] = ptr; //   
	spots->n++;	//  -   
	return ptr;
}

/*
 *    spot.c     ,
 *        
 *       (  )
 */
void spots_center(Spots *spots){
	int i, n;
	double X = 0., Y = 0., N, x0, y0;
	double *xx, *yy, *px, *py;
	Spot **spot;
	n = spots->n; N = (double) n;
	if(n == 0) return;
	xx = malloc(n * sizeof(double)); px = xx;
	yy = malloc(n * sizeof(double)); py = yy;
	spot = spots->spot;
	for(i = 0; i < n; i++, px++, py++, spot++){
				//x0 = (*spot)->c.x; y0 = (*spot)->c.y;
		x0 = (*spot)->c.x - AxisX;
		//y0 = (*spot)->c.y - AxisY;
		y0 = AxisY - (*spot)->c.y; //  ""   
		(*spot)->c.x = x0; (*spot)->c.y = y0;
		*px = x0;
		*py = y0;
		X += *px; Y += *py;
	}
	X /= N; Y /= N;
	DBG(GREEN "spots center: %g, %g\n" OLDCOLOR, X, Y);
	px = xx; py = yy;
	spot = spots->spot;
	for(i = 0; i < n; i++, px++, py++, spot++){
		x0 = *px - X; y0 = *py - Y;
		(*spot)->phi = atan2(y0, x0);
		(*spot)->r = sqrt(x0*x0 + y0*y0);
	}
	free(xx); free(yy);
}

int cmp_spots_r(const void *spot1, const void *spot2){
	return ((*(Spot**)spot1)->r > (*(Spot**)spot2)->r);
}
void sort_spots_by_r(Spots *spots){
	qsort(spots->spot, spots->n, sizeof(Spot*), cmp_spots_r);
}

int cmp_spots_phi(const void *spot1, const void *spot2){
	return ((*(Spot**)spot1)->phi > (*(Spot**)spot2)->phi);
}
void sort_spots_by_phi(Spots *spots){
	qsort(spots->spot, spots->n, sizeof(Spot*), cmp_spots_phi);
}

int cmp_spots_id(const void *spot1, const void *spot2){
	return ((*(Spot**)spot1)->id > (*(Spot**)spot2)->id);
}
void sort_spots_by_id(Spots *spots){
	qsort(spots->spot, spots->n, sizeof(Spot*), cmp_spots_id);
}


/*
 *    R  phi,  
 *   id   
 * < > + 100*< >
 *    100*< >+<     >
 */
void sort_spots(Spots *spots){
	FNAME();
	int i, n;
	int	step = 0,		// "" -    
		num = 0;		//    
	Spot **spot = spots->spot;
	Spots *markers = spots_alloc(2); //    
	double *angles_0; //      "" 
	const int N_RAY_MAX = 2 * N_ANGLES - 1; //   
	const int N_RAYS = 2 * N_ANGLES;
	const int N_ANGLE_MAX = N_ANGLES - 1; // .  
	n = spots->n;
	//   angles_0 (      )
	angles_0 = calloc(N_ANGLES, sizeof(double));
	{
		double ang = aZero;
		for(i = 0; i < N_ANGLES; i++){
			if(ang < -M_PI_2) ang += M_PI;
			if(ang > M_PI_2) ang -= M_PI;
			angles_0[i] = ang;
			ang -= aStep;
			//DBG("angle0[%d]=%g\n", i, angles_0[i]*180/M_PI);
		}
	}
	/*         
	 *        
	 */
	spots_center(spots);
	//    phi       
	{
		sort_spots_by_phi(spots);
		double phi0 = (*spot)->phi, phi, dphi;
		DBG(GREEN "\n\nSort by phi\n\n" OLDCOLOR);
		for(i = 0; i < n; i++, spot++){
			phi = (*spot)->phi;
			dphi = fabs(phi-phi0);
			phi0 = phi;
			/*  2    ,  
			 *   .
			 *      -  */
			if(dphi > PHI_STEP){
				num++;
				if(!step){
					step = 1;
					//(GREEN "NEXT %d, step = %d\n" OLDCOLOR, num, step);
				}else{
					num--;
					DBG(RED "FOUND MARKER\n" OLDCOLOR);
					(*(spot-1))->id = 99 - (*(spot-1))->id;
					DBG("spot[%d]: r=%g, phi=%g\n", (*(spot-1))->id,
						(*(spot-1))->r, (*(spot-1))->phi);
					if(!spots_add(markers, *(spot-1), BY_COPY)){
						bad_circumstances(TOO_MANY_MARKERS);
						goto ret;
					}
				}
			}
			else step = 0;
			(*spot)->id = num;
			//DBG("spot[%d]: r=%g, phi=%g, dphi=%g\n", (*spot)->id, (*spot)->r,
			//	phi, dphi);
		}
	}
	//  
	/* ,       - ,  
	 *      ( -  
	 * ,   - ).
	 */
	{
		double dphi;
		if(markers->n != 2){
			bad_circumstances(NOT_ENOUGH_MARKERS);
			goto ret;
		}
		sort_spots_by_r(markers);
		spot = markers->spot;
		for(i = 0; i < 2; i++)
			DBG(GREEN "marker #%d: r=%g, phi=%g, tmp_id=%d, x=%g, y=%g\n" OLDCOLOR, i,
				spot[i]->r, spot[i]->phi, spot[i]->id, spot[i]->c.x, spot[i]->c.y);
		dphi = spot[0]->phi - spot[1]->phi;
		if(dphi < 0.) dphi += M_2PI;
		DBG("Angle between markers: %g rad\n", dphi);
		if(fabs(dphi - DPHI_MARKERS) > PHI_STEP){
			bad_circumstances(BAD_MARKERS_ANGLE);
			goto ret;
		}
	}
	/*     ,   
	 *     !
	 */
	{
		int mZero = 99 - spot[1]->id; //  ,   
		spot = spots->spot;
		for(i = 0; i < n; i++, spot++){
			int sid;
			if((*spot)->id > N_RAY_MAX){ //   
				sid = 198 - mZero - (*spot)->id;
				if(sid > 99) sid -= N_RAYS;
			}
			else{
				sid = mZero - (*spot)->id; //     
				if(sid < 0) sid += N_RAYS;
			}
			(*spot)->id = sid;
		}
	}
	/*      
	 * y=a+bx, b=tg(slope+alpha_i0)
	 * b = [\sum x_i \sum y_i - n\sum (x_i y_i)] /
	 *        [ (\sum x_i)^2 - n\sum (x_i)^2 ]
	 *
	 *  S1 = \sum x_i, S2 = \sum y_i, S3 = \sum(x_i)^2,
	 * 		S4 = \sum(x_i y_i), 
	 * b = (S1*S2 - n*S4) / (S1*S1 - n*S3)
	 *
	 *  4  Sj  16  (    -)
	 *      ( )  ,
	 *    b     
	 */
	{
		double	*S1, *S2, *S3, *S4, *N, *angles;
		double	x, y;
		int		idx;
		S1 = calloc(N_ANGLES, sizeof(double));
		S2 = calloc(N_ANGLES, sizeof(double));
		S3 = calloc(N_ANGLES, sizeof(double));
		S4 = calloc(N_ANGLES, sizeof(double));
		N  = calloc(N_ANGLES, sizeof(double));
		angles = calloc(N_ANGLES, sizeof(double));
		SlopeAngle = 0.;
		spot = spots->spot;
		for(i = 0; i < n; i++, spot++){
			idx = (*spot)->id;
			if(idx > N_RAY_MAX) continue;
			if(idx > N_ANGLE_MAX) idx -= N_ANGLES; //      
			x = (*spot)->c.x; y = (*spot)->c.y;
			S1[idx] += x;	// \sum x_i
			S2[idx] += y;	// \sum y_i
			S3[idx] += x*x;	// \sum (x_i)^2
			S4[idx] += x*y;	// \sum (x_i y_i)
			N[idx] += 1.;
		}
		for(i = 0; i < N_ANGLES; i++){
			//    i- 
			angles[i] = atan((S1[i]*S2[i] - N[i]*S4[i]) /
				(S1[i]*S1[i] - N[i]*S3[i])) - angles_0[i];
			if(angles[i] < -M_PI_2) angles[i] += M_2PI;
			SlopeAngle += angles[i];
			//DBG("angle[%d] = %g\n", i, angles[i]*180/M_PI);
		}
		SlopeAngle /= (double)N_ANGLES; // 
		DBG(RED "\n\nSlope Angle: %g degr\n" OLDCOLOR, SlopeAngle * 180. / M_PI);
		free(S1); free(S2); free(S3); free(S4); free(N); free(angles);
	}
	//    r    
	{
		sort_spots_by_r(spots);
		spot = spots->spot;
		double r0 = (*spot)->r, r;
		// rtres - 1/7     
		double rtres = (spot[n-1]->r - r0) / ((double)N_RINGS - 1.) / 7.;
		num = 0;
		DBG(RED "\n\nsort by r\n\n" OLDCOLOR);
		for(i = 0; i < n; i++, spot++){
			r = (*spot)->r;
			if(fabs(r-r0) > rtres){
				num+=100;
				//DBG(GREEN "NEXT, num=%d\n" OLDCOLOR, num);
			}
			(*spot)->id += num;
		//	DBG("spot[%d]: r=%g, phi=%g, dr=%g\n", (*spot)->id, r,
		//		(*spot)->phi, r - r0);
			r0 = r;
		}
		if((num/100 + 1) > N_RINGS)
			bad_circumstances(TOO_MANY_RINGS);
	}
	/*
	 *          , 
	 * ,     .
	 */


	/*        
	 *    :  +   . .
	 * (\beta == SlopeAngle)
	 * x^0_i = -x_i\cos\beta - y_i\sin\beta;
	 * y^0_i = -x_i\sin\beta + y_i\cos\beta
	 */
	{
		double cosSA = cos(SlopeAngle);
		double sinSA = sin(SlopeAngle);
		double x,y;
		spot = spots->spot;
		for(i = 0; i < n; i++, spot++){
			x = (*spot)->c.x; y = (*spot)->c.y;
			(*spot)->xC = x*cosSA + y*sinSA;
			(*spot)->yC = -x*sinSA + y*cosSA;
		}
	}
/* debug ======>

	sort_spots_by_id(spots);
	spot = spots->spot;
	DBG(RED "\n\nFinal sort\n\n" OLDCOLOR);
	for(i = 0; i < n; i++, spot++){
		DBG("spot[%d]: xc=%.1f, yc=%.1f\n", (*spot)->id,
			(*spot)->c.x, (*spot)->c.y);
	}
 <===== */
ret:
	spots_free(&markers);
	free(angles_0);
}

#else // LEPTONICA_FOUND not set
Spots* spots_alloc(int n){
	LEPTERR;
	printf("%d\n", n); return NULL;
}
Spot* spots_add(Spots *spots, Spot *spot, int copy){
	LEPTERR;
	if(spots || spot)printf("%d\n", copy);
	return NULL;
}
void spots_free(Spots **spots){
	LEPTERR;
	if(spots) return;
}
void sort_spots(Spots *spots){
	LEPTERR;
	if(spots) return;
}
#endif // LEPTONICA_FOUND
